/* CS255 iButton extra credit starter file 
 * This file implements the Host side code...
 *
 * Created: 1 Mar 2000  by Michael Kim
 *
 * This class eventually will be a handler for the iButton.
 * It instantiates listeners and will give commands to and 
 * receive response from the iButton.
 * 
 */ 

import com.ibutton.oc.*;
import com.ibutton.oc.CommandAPDU;
import opencard.core.terminal.*;
import com.ibutton.opt.jibletselect.Selector;
import java.io.*;
import java.util.*;
import java.util.Random.*;
import java.io.File;

public class CS255_iButton_Host implements JibMultiListener
{
    // BEGIN INSTRUCTION DECLARATIONS
    public static final byte CS255_IBUTTON_CLA = (byte)0x80;
    public static final byte CS255_IBUTTON_INS_SET_PIN = (byte)0;
    public static final byte CS255_IBUTTON_INS_GENERATE_KEYS = (byte)1;
    public static final byte CS255_IBUTTON_INS_GET_PUBLIC_KEY = (byte)2;
    public static final byte CS255_IBUTTON_INS_SIGN_MESSAGE = (byte)3;
    public static final byte CS255_IBUTTON_INS_IS_FIRST_TIME = (byte)4;
    //   END INSTRUCTION DECLARATIONS

    //The object that will be used to communicate with the iButtons.
    protected JibMultiFactory factory;

    //The name (AID) of the iButton applet used by this host.
    private String appletName;
    private String appletPath;

    byte[] currentPIN, newPIN;
    private String guessedPIN;


    /**
     * Sets up the listener for iButton inserted and iButton
     * removed events.
     *
     * @param appletPath  the path to the applet that should
     *                    be loaded into the iButtons.
     * @param appletName  the name of the applet that should
     *                    be loaded into the iButtons.
     */
    public CS255_iButton_Host(String appletPath, String appletName, String passPhrase)
    {
        this.appletName = appletName;
        this.appletPath = appletPath;

	this.guessedPIN = new String(passPhrase);
        //*************************************
        //* Add any initialization code here. *
        //*************************************

        factory = new JibMultiFactory();
        //Add the listener that will be notified
        //when an iButton is inserted or removed.
        factory.addJiBListener(this);
        //Begin polling for iButtons.
        try
        {
            factory.startPolling(true);
        }
        catch(CardTerminalException cte)
        {
            cte.printStackTrace();
        }
    }

    /**
     * Adds the specified iButton listener to receive
     * iButton inserted and removed events.
     *
     * @param l  the listener.
     */
    public void addJiBListener(JibMultiListener l)
    {
        factory.addJiBListener(l);
    }

    /**
     * Removes the specified iButton listener so that
     * the listener no longer receives iButton inserted
     * and removed events.
     *
     * @param l  the listener.
     */
    public void removeJiBListener(JibMultiListener l)
    {
        factory.removeJiBListener(l);
    }

    /**
     * Called when an iButton gets inserted.
     *
     * @param event  the insertion event.
     */
    public void iButtonInserted(JibMultiEvent event)
    {
        //The master pin for this iButton.
        //If a password is set on the iButton you must
        //provide that here.
        String password = new String(guessedPIN);

        factory.setSlot(event.getSlotID());

        try
        {
            //Try to select the applet.
            Selector appletSelector = new Selector(false);
            if(appletSelector.select(appletPath,appletName + ".jib",
                                     appletName, event.getChannel(),
                                     password))
            {

                //*****************************************
                //* Insert any code to be done when       *
                //* an iButton is inserted here.          *
                //*****************************************
		
		System.out.println("iButton inserted.");
                
                // Turn on Applet Garbage Collection
                // Because your code will do BigInteger calls, 
		// it is essential to have this feature enabled.
		factory.setAppletGCMode(1);
                System.out.println("Applet GC enabled.");
                               
                // CS255: Extra Credit
                
                // If this is the first time the iButton has been used
		if(is_first_timeDispatch() == 1) 
                {
                    set_pinDispatch();
                    generate_keysDispatch();
                }
		                	
            }
        }
        catch(CardTerminalException cte)
        {
            System.err.println("ERROR:  CardTerminalException occurred while communicating with iButton.");
        }
        catch(FileNotFoundException fnfe)
        {
            System.err.println("Unable to find file: " + appletName + ".jib");
        }
        catch(IOException ioe)
        {
            System.err.println("Unable to read file: " + appletName + ".jib");
        }
        catch(Exception e)
        {
            //Exceptions that occur in iButtonInserted events
            //will be drained in the OpenCard internals if we
            //don't catch them here.
            System.err.println("Exception in iButtonInserted:");
            e.printStackTrace();
        }
    }

    /**
     * Called when an iButton is removed.
     *
     * @param event  the removal event.
     */
    public void iButtonRemoved(JibMultiEvent event)
    {
        try
        {
            //*****************************************
            //* Insert any code to be done when       *
            //* an iButton is removed here.           *
            //*****************************************
            
            System.out.println("iButton removed.");
        }
        catch(Exception e)
        {
            //Exceptions that occur in iButtonRemoved events
            //will be drained in the OpenCard internals if we
            //don't catch them here.
            System.err.println("Exception in iButtonRemoved:");
            e.printStackTrace();
        }
    }

    /**
     * Will interpret a status word into a string describing
     * the meaning of that status word..
     *
     * @param sw  the status word to decode.
     *
     * @return the string describing the status.
     */
    private String decodeSW(int sw)
    {
        switch (sw)
        {
            case 0x6100:
                return "Response Bytes Remaining";
            case 0x6301:
                return "Success Packet";
            case 0x6400:
                return "Insufficient Memory";
            case 0x6681:
                return "Bad Master PIN";
            case 0x6700:
                return "Wrong Length";
            case 0x6901:
                return "Invalid AID Length";
            case 0x6902:
                return "Invalid API Version";
            case 0x6903:
                return "Invalid Password";
            case 0x6904:
                return "Invalid Signature Length";
            case 0x6905:
                return "Hash Corruption";
            case 0x6906:
                return "Hash Failure";
            case 0x6982:
                return "Invalid Signature";
            case 0x6984:
                return "Data Invalid";
            case 0x6985:
                return "Conditions Of Use Not Satisfied";
            case 0x6A80:
                return "Wrong Data";
            case 0x6A81:
                return "Function Not Supported";
            case 0x6A82:
                return "Unable to Select Applet";
            case 0x6A84:
                return "Class Length Overrun";
            case 0x6A86:
                return "Invalid Loader Command";
            case 0x6A87:
                return "Incomplete Packet";
            case 0x6B00:
                return "Incorrect Parameters (P1,P2)";
            case 0x6C00:
                return "Correct Expected Length";
            case 0x6D00:
                return "INS Value Not Supported";
            case 0x6F00:
                return "Uncaught Exception";
            case 0x8450:
                return "Unable to Find Applet";
            case 0x8453:
                return "Unable to Select Applet";
            case 0x9000:
                return "Success";
            default:
                return "Unrecognized SW";
        }
    }

    // BEGIN INSTRUCTION DISPATCHER FUNCTIONS
    public int is_first_timeDispatch() throws CardTerminalException
    {
        ResponseAPDU response;

        response = factory.sendAPDU(new CommandAPDU((byte)CS255_IBUTTON_CLA, (byte)CS255_IBUTTON_INS_IS_FIRST_TIME, (byte)0x0, (byte)0x0, null /*add data to send here*/, (byte)0));
        String responseStr = decodeSW(response.sw());
        System.out.println(responseStr);

	// For now assume that it is always needing to be initialized
        return(1);  
    }

    public int is_first_timeDispatch(int slotNumber) throws CardTerminalException
    {
        factory.setSlot(slotNumber);
	    return(is_first_timeDispatch());
    }
  
    public void set_pinDispatch() throws CardTerminalException
    {
        ResponseAPDU response;
        
        response = factory.sendAPDU(new CommandAPDU((byte)CS255_IBUTTON_CLA, (byte)CS255_IBUTTON_INS_SET_PIN, (byte)0x0, (byte)0x0, null /*add data to send here*/, (byte)0));
        String responseStr = decodeSW(response.sw());
        System.out.println(responseStr);
    }

    public void set_pinDispatch(int slotNumber) throws CardTerminalException
    {
        factory.setSlot(slotNumber);
        set_pinDispatch();
    }

    public void generate_keysDispatch() throws CardTerminalException
    {
        ResponseAPDU response;

        response = factory.sendAPDU(new CommandAPDU((byte)CS255_IBUTTON_CLA, (byte)CS255_IBUTTON_INS_GENERATE_KEYS, (byte)0x0, (byte)0x0, null /*add data to send here*/, (byte)0));
        String responseStr = decodeSW(response.sw());
        System.out.println(responseStr);
    }

    public void generate_keysDispatch(int slotNumber) throws CardTerminalException
    {
        factory.setSlot(slotNumber);
        generate_keysDispatch();
    }

    public void get_public_keyDispatch() throws CardTerminalException
    {
        ResponseAPDU response;

        response = factory.sendAPDU(new CommandAPDU((byte)CS255_IBUTTON_CLA, (byte)CS255_IBUTTON_INS_GET_PUBLIC_KEY, (byte)0x0, (byte)0x0, null /*add data to send here*/, (byte)0));
        String responseStr = decodeSW(response.sw());
        System.out.println(responseStr);
    }

    public void get_public_keyDispatch(int slotNumber) throws CardTerminalException
    {
        factory.setSlot(slotNumber);
        get_public_keyDispatch();
    }

    public void sign_messageDispatch() throws CardTerminalException
    {
        ResponseAPDU response;

        response = factory.sendAPDU(new CommandAPDU((byte)CS255_IBUTTON_CLA, (byte)CS255_IBUTTON_INS_SIGN_MESSAGE, (byte)0x0, (byte)0x0, null /*add data to send here*/, (byte)0));
        String responseStr = decodeSW(response.sw());
        System.out.println(responseStr);
    }

    public void sign_messageDispatch(int slotNumber) throws CardTerminalException
    {
        factory.setSlot(slotNumber);
        sign_messageDispatch();
    }
    //   END INSTRUCTION DISPATCHER FUNCTIONS

    // BEGIN CONVENIENCE FUNCTIONS
    public static void intToByteArray(byte[] outArray, int start, int value)
    {
        //BigEndian
        outArray[start]     = (byte)((value & 0xFF000000) >>> 24);
        outArray[start + 1] = (byte)((value & 0x00FF0000) >>> 16);
        outArray[start + 2] = (byte)((value & 0x0000FF00) >>>  8);
        outArray[start + 3] = (byte) (value & 0x000000FF);
    }

    public static int intFromByteArray(byte[] inArray, int start)
    {
        //BigEndian
        return (((int)inArray[start    ] << 24) & 0xFF000000 |
                ((int)inArray[start + 1] << 16) & 0x00FF0000 |
                ((int)inArray[start + 2] <<  8) & 0x0000FF00 |
                ((int)inArray[start + 3]      ) & 0x000000FF);
    }

    public static void intArrayToByteArray(byte[] outArray, int start, int[] value)
    {
        //BigEndian
        for(int i = 0; i < value.length; i++)
        {
            outArray[start + (i * 4)]     = (byte)((value[i] & 0xFF000000) >>> 24);
            outArray[start + (i * 4) + 1] = (byte)((value[i] & 0x00FF0000) >>> 16);
            outArray[start + (i * 4) + 2] = (byte)((value[i] & 0x0000FF00) >>>  8);
            outArray[start + (i * 4) + 3] = (byte) (value[i] & 0x000000FF);
        }
    }

    public static int[] intArrayFromByteArray(byte[] inArray, int start, int length)
    {
        int[] ia = new int[length / 4];
        //BigEndian
        for(int i = 0; i < length / 4; i++)
        {
            ia[i] = (((int)inArray[start + (i * 4)    ] << 24) & 0xFF000000 |
                     ((int)inArray[start + (i * 4) + 1] << 16) & 0x00FF0000 |
                     ((int)inArray[start + (i * 4) + 2] <<  8) & 0x0000FF00 |
                     ((int)inArray[start + (i * 4) + 3]      ) & 0x000000FF);
        }
        return ia;
    }
    //   END CONVENIENCE FUNCTIONS

    public static void main(String[] args)
    {
        //The following path has been auto-generated by
        //the Project Wizard.  If the project is moved
        //to a different directory, you MUST update this
        //path accordingly.
        new CS255_iButton_Host("c:\\windows\\desktop\\CS255_iButton\\iButton\\Classes\\", "CS255_iButton", "" /*pass phrase*/);
    }
}









