/* CS255 iButton extra credit starter file 
 * This file implements the Applet side code...
 *
 * Created: 1 Mar 2000  by Michael Kim
 *
 * This applet is responsible for generating, storing signing keys.
 * On demand, it should also give the public verifying key.
 * It should also hash and sign things using the signing keys.
 * 
 */
 
  
import javacard.framework.*;
import javacard.framework.Util;
import com.dalsemi.system.*;
import opencard.opt.security.*;

public class CS255_iButton_Applet extends Applet
{
    // BEGIN INSTRUCTION DECLARATIONS
    public static final byte CS255_IBUTTON_CLA = (byte)0x80;
    public static final byte CS255_IBUTTON_INS_SET_PIN = (byte)0;
    public static final byte CS255_IBUTTON_INS_GENERATE_KEYS = (byte)1;
    public static final byte CS255_IBUTTON_INS_GET_PUBLIC_KEY = (byte)2;
    public static final byte CS255_IBUTTON_INS_SIGN_MESSAGE = (byte)3;
    public static final byte CS255_IBUTTON_INS_IS_FIRST_TIME = (byte)4;
    //   END INSTRUCTION DECLARATIONS

    //The parameters used to select an applet.
    final static private byte SELECT_CLA = (byte)0x00;
    final static private byte SELECT_INS = (byte)0xA4;
    
    //The Maximum mumber of bytes we will send in a single apdu.
    final static private short MAX_SEND_LENGTH = (short)1000;

    private OwnerPIN userPIN = null;
    private short    pinSize;
    
    private int lastPINCheckTime;
    private int firstTime;

    public byte[] apduData;

    public CS255_iButton_Applet()
    {
        //Register this applet with the JCRE
        register();

        //**************************************
        //* Add any initialization code here.  *
        //**************************************
        
        //firstTime set to 1.
        //Create a new OwnerPin with a max try limit of 5 and
	//max length of 8
    	//no default PIN is set.
	firstTime = 1;
	pinSize = 0;
	byte[] pin = new byte[0];

	/* For an explanation of what these lines do, see checkPin */
        userPIN = new javacard.framework.OwnerPIN((byte)0x05,(byte)0x08);
        userPIN.updateAndUnblock(pin, (short)0, (byte)0);
    }

    public static void install(APDU apdu)
    {
        new CS255_iButton_Applet();
    }

    public void process(APDU apdu)
    {
        byte[] buffer = apdu.getBuffer();

        //Determine if the applet is being selected.
        if((buffer[ISO.OFFSET_CLA] == SELECT_CLA) &&
           (buffer[ISO.OFFSET_INS] == SELECT_INS))
        {
            //***********************************
            //* Add any code to be executed on  *
            //* applet selection here.          *
            //***********************************

	    // CS255: There shouldn't need to be any code here
	    // unless you so desired...

            return;
        }

        apduData = new byte[buffer[ISO.OFFSET_LC] & 0x0FF];
        short apduDataOffset = 0;

        //Read in the entire APDU.
        short bytesRead = apdu.setIncomingAndReceive();
        //Loop until all bytes have been read.
        while (bytesRead > 0)
        {
            Util.arrayCopyNonAtomic(buffer, ISO.OFFSET_CDATA, apduData, apduDataOffset, bytesRead);
            apduDataOffset += bytesRead;
            bytesRead = apdu.receiveBytes(ISO.OFFSET_CDATA);
        }
        //Prepare the apdu for sending.
        apdu.setOutgoing();
        apdu.setOutgoingLength(MAX_SEND_LENGTH);

        //Check for a valid CLA.
        if(buffer[ISO.OFFSET_CLA] != CS255_IBUTTON_CLA)
        {
            ISOException.throwIt(ISO.SW_CLA_NOT_SUPPORTED);
        }
        else
        {
            //Call the appropriate dispatch method for the given INS.
            switch (buffer[ISO.OFFSET_INS])
            {            
                case CS255_IBUTTON_INS_IS_FIRST_TIME:
                    is_first_timeDispatch(apdu, buffer[ISO.OFFSET_P1], buffer[ISO.OFFSET_P2]);
                    break;              
                case CS255_IBUTTON_INS_SET_PIN:
                    set_pinDispatch(apdu, buffer[ISO.OFFSET_P1], buffer[ISO.OFFSET_P2]);
                    break;
                case CS255_IBUTTON_INS_GENERATE_KEYS:
                    generate_keysDispatch(apdu, buffer[ISO.OFFSET_P1], buffer[ISO.OFFSET_P2]);
                    break;
                case CS255_IBUTTON_INS_GET_PUBLIC_KEY:
                    get_public_keyDispatch(apdu, buffer[ISO.OFFSET_P1], buffer[ISO.OFFSET_P2]);
                    break;
                case CS255_IBUTTON_INS_SIGN_MESSAGE:
                    sign_messageDispatch(apdu, buffer[ISO.OFFSET_P1], buffer[ISO.OFFSET_P2]);
                    break;
                default:
                    ISOException.throwIt(ISO.SW_INS_NOT_SUPPORTED);
            }
        }
    }

    protected void sendByteArray(APDU apdu, byte[] data)
    {
        short offset = 0;
        while((data.length - offset) > MAX_SEND_LENGTH)
        {
            apdu.sendBytesLong(data, offset, MAX_SEND_LENGTH);
            offset += MAX_SEND_LENGTH;
        }
        apdu.sendBytesLong(data, offset, (short)(data.length - offset));
    }

    // BEGIN INSTRUCTION DISPATCHER FUNCTIONS
    public void is_first_timeDispatch(APDU apdu, byte p1, byte p2) 
    {
        /* This needs to sendByteArray corresponding to whether or not
	   the iButton needs to set a PIN and generate a key pair */

	/* Has this iButton been initialized? */

    }

    public void set_pinDispatch(APDU apdu, byte p1, byte p2) 
    {
        /* The first thing you do when you have a new iButton
	   is set the pin, and this also sets firstTime to 0. */
	firstTime = 0;
        
	/* PIN setting code here */
        
    }

    public void generate_keysDispatch(APDU apdu, byte p1, byte p2) 
    {
	/* add code to generate the DSA key pairs here */
    }

    public void get_public_keyDispatch(APDU apdu, byte p1, byte p2) 
    {
        /* this code should sendByteArray corresponding 
	   to the public verifying key */

    }

    public void sign_messageDispatch(APDU apdu, byte p1, byte p2) 
    {
        /* Passed into this function will be a byte array.
	   The array will have the message contents which should be:
	   hashed, then signed, then send the signature back */

    }
    //   END INSTRUCTION DISPATCHER FUNCTIONS

    // BEGIN CONVENIENCE FUNCTIONS
    
 /************************************************************************\
            Function:    checkPin()
            Parameters:  none
            Returns:     void
            Purpose:     This function checks to see if the PIN passed in from
                    the host matches this applet's internal PIN.  If not, it
                    throws an exception.  The first byte should be the length
                    of the PIN supplied, followed by length bytes of PIN data.
                    A host application has 5 tries to get the correct PIN.  If
                    an incorrect PIN is supplied 5 times in a row, the PIN
                    becomes blocked and cannot be used again (even with the 
                    correct PIN value) for a duration of 30 minutes.  After
                    this time period has elapsed, the host may once again
                    attempt to send a correct PIN.
    \************************************************************************/
    void checkPin()
    {
        int currentTime = Clock.getClock();
        if (userPIN.getTriesRemaining() == 0)
        {
            if ( (currentTime - lastPINCheckTime) > 1800)
            	userPIN.resetAndUnblock();
        }

      	lastPINCheckTime = currentTime;
        
      	if ( (apduData[0] != pinSize) | (!userPIN.check(apduData,(short)1,apduData[0])) )
           	throw new ISOException(ISO.SW_SECURITY_STATUS_NOT_SATISFIED);
    }
  
    public static void intToByteArray(byte[] outArray, int start, int value)
    {
        //BigEndian
        outArray[start]     = (byte)((value & 0xFF000000) >>> 24);
        outArray[start + 1] = (byte)((value & 0x00FF0000) >>> 16);
        outArray[start + 2] = (byte)((value & 0x0000FF00) >>>  8);
        outArray[start + 3] = (byte) (value & 0x000000FF);
    }

    public static int intFromByteArray(byte[] inArray, int start)
    {
        //BigEndian
        return (((int)inArray[start    ] << 24) & 0xFF000000 |
                ((int)inArray[start + 1] << 16) & 0x00FF0000 |
                ((int)inArray[start + 2] <<  8) & 0x0000FF00 |
                ((int)inArray[start + 3]      ) & 0x000000FF);
    }

    public static void intArrayToByteArray(byte[] outArray, int start, int[] value)
    {
        //BigEndian
        for(int i = 0; i < value.length; i++)
        {
            outArray[start + (i * 4)]     = (byte)((value[i] & 0xFF000000) >>> 24);
            outArray[start + (i * 4) + 1] = (byte)((value[i] & 0x00FF0000) >>> 16);
            outArray[start + (i * 4) + 2] = (byte)((value[i] & 0x0000FF00) >>>  8);
            outArray[start + (i * 4) + 3] = (byte) (value[i] & 0x000000FF);
        }
    }

    public static int[] intArrayFromByteArray(byte[] inArray, int start, int length)
    {
        int[] ia = new int[length / 4];
        //BigEndian
        for(int i = 0; i < length / 4; i++)
        {
            ia[i] = (((int)inArray[start + (i * 4)    ] << 24) & 0xFF000000 |
                     ((int)inArray[start + (i * 4) + 1] << 16) & 0x00FF0000 |
                     ((int)inArray[start + (i * 4) + 2] <<  8) & 0x0000FF00 |
                     ((int)inArray[start + (i * 4) + 3]      ) & 0x000000FF);
        }
        return ia;
    }
    //   END CONVENIENCE FUNCTIONS
}
